/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "message_box.h"
#include "modules.h"
#include "misc.h"
#include "tcl_scripting.h"

#ifdef USE_TK
#include <tk.h>
#endif

#define MODULE_TCL_VERSION "0.5.5-Gueneveth"

// Initalisation of module
TclSessions tclSess;
TclLoadDialog *loadDialog;
QTimer* events_loop;
PubDirID tclPubDirList;

extern "C" int tcl_scripting_init()
{
    kdebug("Initializing TCL module.");
    loadDialog = new TclLoadDialog();
    new TclSession(TRUE);

    QString path = ggPath(".module_tcl_scripts");

    QFile file(path);
    if ( file.open( IO_ReadOnly ) )
    {
        QTextStream stream(&file);
        QString line;
        while (!stream.atEnd())
        {
            line = stream.readLine();
            if (!line.isEmpty())
            {
                TclSession *tsess = new TclSession();
                if (tsess->load(line) == TCL_ERROR)
                {
                    tsess->error(QString::fromUtf8(Tcl_GetStringResult(tsess->tclInt())), TRUE);
                    delete tsess;
                }
                else
                {
                    QString name = QStringList::split("/", line).last();
                    if (!loadDialog->loaded(name))
                    {
                        loadDialog->list()->insertItem(name);
                        for ( uint i = 0; i < loadDialog->lList()->count(); i++ )
                        {
                            if (loadDialog->lList()->text(i) == name)
                                loadDialog->lList()->removeItem(i);
                        }
                        for ( uint i = 0; i < loadDialog->gList()->count(); i++ )
                        {
                            if (loadDialog->gList()->text(i) == name)
                                loadDialog->gList()->removeItem(i);
                        }
                    }
                }
            }
        }
        file.close();
    }

    events_loop = new QTimer(0);
    QObject::connect(events_loop, SIGNAL(timeout()), tclSess[0], SLOT(processOneEvent()) );
    events_loop->start( 1, FALSE );

    return 0;
}

extern "C" void tcl_scripting_close()
{
    kdebug("Closing TCL module.\n");

    events_loop->stop();
    QObject::disconnect(events_loop, SIGNAL(timeout()), tclSess[0], SLOT(processOneEvent()) );
    delete events_loop;

    QString path = ggPath(".module_tcl_scripts");
    QFile file(path);
    if ( file.open( IO_WriteOnly ) )
    {
        QTextStream stream(&file);
        TclSessions::Iterator it = tclSess.begin();
        for ( ++it ; it != tclSess.end(); it++ )
        {
            stream << it.data()->scriptName() << endl;
        }
        file.close();
    }
    else
    {
        QMessageBox::warning( 0, QObject::tr("Warning!"), QObject::tr("Unable to open file %1 for writting. Can't save scripts list.").arg(path) );
    }
    for ( int i = tclSess.count() - 1 ; i > -1 ; i-- )
    {
        tclSess[i]->tclEvtHnd()->unloadModule();
        delete tclSess[i];
    }

    delete loadDialog;
}


// Implementation of TclSession class

TclSession::TclSession(bool main)
{
    kdebugf();

    int i = 0;
    while (tclSess.contains(i))
        i++;

    guiEnabled = FALSE;
    setID(i);
    tclSess[i] = this;
    idPtr = new int(i);

    tclrun = TRUE;
    isMain = FALSE;
    tclEventHandler = new TclEventHandler(this);
    start();
 
    if (main)
    {
        isMain = TRUE;
        Tcl_cons = new TclConsole();
        tclMenu = new QPopupMenu(kadu);
        addMenu();
        if (load() == TCL_ERROR)
        {
            error(QString::fromUtf8(Tcl_GetStringResult(tcl_int)), TRUE);
        }
    }
}

TclSession::~TclSession()
{
    kdebugf();

    for ( QMap<QString,int>::Iterator it = tclMenuBar.begin(); it != tclMenuBar.end(); it++ )
        kadu->menuBar()->removeItem(*it);

    stop();
    if (isMain)
    {
        delMenu();
        delete Tcl_cons;
        delete tclMenu;
    }
    delete tclEventHandler;
    delete idPtr;
    tclSess.remove(sessID);
}

void TclSession::openConsole()
{
    kdebugf();
    if (Tcl_cons->isShown())
    {
        Tcl_cons->raise();
    }
    else
    {
        Tcl_cons->show();
    }
    Tcl_cons->getCmdLine()->setFocus();
}

void TclSession::openLoadDialog()
{
    kdebugf();
    loadDialog->show();
}

void TclSession::addMenu()
{
    kdebugf();
    tclMenu->insertItem(icons_manager.loadIcon("ResendUserlist"), tr("Load dialog"), this, SLOT(openLoadDialog()), 0, -1, 0);
    tclMenu->insertItem(icons_manager.loadIcon("AutoSendMessage"), tr("Open console"), this, SLOT(openConsole()), 0, -1, 1);
    tclMenu->insertSeparator(2);
    if (!config_file.readEntry("TCL", "hidemenu").compare("true"))
    {
        tclMenu->insertItem(icons_manager.loadIcon("SendMessage"), tr("Show TCL menu"), this, SLOT(showTclMenu()), 0, -1, 3);
        menu_id = kadu->mainMenu()->insertItem(tr("&TCL module"), tclMenu, -1, 16);
    }
    else
    {
        tclMenu->insertItem(icons_manager.loadIcon("CancelMessage"), tr("Hide TCL menu"), this, SLOT(hideTclMenu()), 0, -1, 3);
        menu_id = kadu->menuBar()->insertItem(tr("&TCL module"), tclMenu, -1, 1);
    }
    tclMenu->insertItem(icons_manager.loadIcon("History"), tr("About module"), this, SLOT(aboutModule()), 0, -1, 4);
}

void TclSession::aboutModule()
{
    kdebugf();
    QString info;
    info += tr("<b>Module TCL v%1</b><br>");
    info += tr("This module allows to use TCL scripting language under Kadu.") + "<br>";
    info += tr("Home page of project is:") + " ";
    info += "<a href=\"http://scripts.one.pl/tcl4kadu\">http://scripts.one.pl/tcl4kadu</a><br><br>";
    info += tr("Author is Pawel (Googie) Salawa") + " ";
    info += "(<a href=\"mailto:boogie@scripts.one.pl\">boogie@scripts.one.pl</a>).";
    MessageBox::msg(info.arg(MODULE_TCL_VERSION));
}

void TclSession::menuItemSetEnabled(int idx, bool enabled)
{
    kdebugf();
    tclMenu->setItemEnabled(tclMenu->idAt(idx), enabled);
}

void TclSession::delMenu()
{
    kdebugf();
    int itemId = -1;
    for ( uint i = 0; i <= kadu->mainMenu()->count(); i++ )
    {
        if ( ! kadu->mainMenu()->text(kadu->mainMenu()->idAt(i)).compare(tr("&TCL module")) )
        {
            itemId = kadu->mainMenu()->idAt(i);
        }
    }
    if (itemId != -1)
        kadu->mainMenu()->removeItem(menu_id);
    else
        kadu->menuBar()->removeItem(menu_id);

    TclPopups::Iterator it;
    for ( it = tclPopups.begin(); it != tclPopups.end(); it++ )
    {
        kadu->menuBar()->removeItem(it.data().id);
    }
}

void TclSession::hideTclMenu()
{
    kdebugf();
    kadu->menuBar()->removeItem(menu_id);
    menu_id = kadu->mainMenu()->insertItem(tr("&TCL module"), tclMenu, -1, 17);
    tclMenu->removeItemAt(3);
    tclMenu->insertItem(icons_manager.loadIcon("SendMessage"), tr("Show TCL menu"), this, SLOT(showTclMenu()), 0, -1, 3);
    config_file.writeEntry("TCL", "hidemenu", "true");
}

void TclSession::showTclMenu()
{
    kdebugf();
    kadu->mainMenu()->removeItem(menu_id);
    menu_id = kadu->menuBar()->insertItem(tr("&TCL module"), tclMenu, -1, 1);
    tclMenu->removeItemAt(3);
    tclMenu->insertItem(icons_manager.loadIcon("CancelMessage"), tr("Hide TCL menu"), this, SLOT(hideTclMenu()), 0, -1, 3);
    config_file.writeEntry("TCL", "hidemenu", "false");
}

bool TclSession::start()
{
    kdebugf();
    QString arg = qApp->argv()[0];

    // Starting of TCL interpreter
    Tcl_FindExecutable(arg.latin1());
    tcl_int = Tcl_CreateInterp();
    kdebug("Tcl_Init(): Initializing TCL interpreter.\n");
    if (Tcl_Init(tcl_int) == TCL_ERROR)
    {
        error(tr("Error while initializing TCL intepreter:\n%1").arg(Tcl_GetStringResult(tcl_int)));
        return TCL_ERROR;
    }

    // Adding new commands (see commands.cpp)
    addTclCommands();
    
    // Setting command line arguments
    // argv0
    Tcl_SetVar(tcl_int, "argv0", arg.latin1(), TCL_GLOBAL_ONLY);

    // argv
    arg = qApp->argv()[1];
    for ( int i = 2; i < qApp->argc(); i++ )
    {
        arg.append(" ");
        arg.append(qApp->argv()[i]);
    }
    Tcl_SetVar(tcl_int, "argv", arg.latin1(), TCL_GLOBAL_ONLY);

    // argc
    int c = qApp->argc() - 1;
    Tcl_SetVar(tcl_int, "argc", (char*)QString::number(c).latin1(), TCL_GLOBAL_ONLY);

    // Setting some useful global variables
    QString dir = ggPath("");

    QStringList l = QStringList::split("-", MODULE_TCL_VERSION);
    Tcl_SetVar(tcl_int, "module(version)", l[0].latin1(), TCL_GLOBAL_ONLY);
    Tcl_SetVar(tcl_int, "module(codename)", l[1].latin1(), TCL_GLOBAL_ONLY);

    Tcl_SetVar(tcl_int, "kadu(datadir)", dataPath().latin1(), TCL_GLOBAL_ONLY);
    Tcl_SetVar(tcl_int, "kadu(configdir)", dir.latin1(), TCL_GLOBAL_ONLY);
    Tcl_SetVar(tcl_int, "kadu(version)", QString(VERSION).latin1(), TCL_GLOBAL_ONLY);

    Tcl_SetVar(tcl_int, "gg(nick)", config_file.readEntry("General", "Nick"), TCL_GLOBAL_ONLY);
    Tcl_SetVar(tcl_int, "gg(uin)", config_file.readEntry("General", "UIN"), TCL_GLOBAL_ONLY);

    return TCL_OK;
}

int TclSession::load(QString file)
{
    kdebugf();
    if (tclrun && QFile::exists(file))
    {
        QStringList tmp = QStringList::split("/", file);
        tmp.remove(tmp.last());
        QString path = "/" + tmp.join("/");
        Tcl_SetVar(tcl_int, "module(scriptpath)", path.latin1(), TCL_GLOBAL_ONLY);

        int status = Tcl_EvalFile(tcl_int, file);
        scriptFile = file;
        tclSess[0]->console()->context()->insertItem(QStringList::split("/", file).last());

        return status;
    }
    else
    {
        return TCL_ERROR;
    }
}

int TclSession::load()
{
    kdebugf();
    QString path = ggPath("default.tcl");
    if (QFile::exists(path))
        return load(path);
    else
        return TCL_OK;
}

void TclSession::stop()
{
    kdebugf();
    if (tclrun)
    {
        killAllTimers();
        Tcl_DeleteInterp(tcl_int);

        int i = 0;
        QString file = QStringList::split("/", scriptName()).last();
        while ( i < tclSess[0]->console()->context()->count()
            && file != tclSess[0]->console()->context()->text(i) )
            i++;

        if (i < tclSess[0]->console()->context()->count())
            tclSess[0]->console()->context()->removeItem(i);

        tclrun = FALSE;
    }
}

bool TclSession::isrun()
{
    kdebugf();
    return tclrun;  
}

void TclSession::error(QString cmd, bool line)
{
    kdebug("void TclSession::error(QString, bool): \"%s\", in line %d.\n", cmd.latin1(), tcl_int->errorLine);
    QString error = QString(tr("Error in TCL script:\n%1")).arg(cmd);
    if (line)
    {
        error.append("\n");
        error.append(tr("In line:"));
        error.append(" ");
        error.append(QString::number(tcl_int->errorLine));
    }
    MessageBox::msg(error);
}

void TclSession::handlePopup(int id)
{
    kdebugf();
    kdebug("Handle popup %d, command code: %s\n", id, tclPopupCmds[id].latin1());
    QCString utf = tclPopupCmds[id].utf8();
    Tcl_Obj* result = Tcl_NewStringObj((const char* const)utf, utf.length());
    int ret = Tcl_Eval(tcl_int, Tcl_GetString(result));
    if (ret)
    {
        error(QString::fromUtf8(Tcl_GetStringResult(tcl_int)), TRUE);
    }
}

void TclSession::handleUserMenu(int id)
{
    kdebugf();
    kdebug("Handle user menu %d, command code: %s\n", id, tclMenuCmds[id].latin1());
    int ret = Tcl_Eval(tcl_int, tclMenuCmds[id]);
    if (ret)
    {
        error(QString::fromUtf8(Tcl_GetStringResult(tcl_int)), TRUE);
    }
}

QString TclSession::scriptName()
{
    kdebugf();
    return scriptFile;
}

void TclSession::killAllTimers()
{
    kdebugf();
    QValueList<timerID>::Iterator it;
    for ( it = tclTimers.begin(); it != tclTimers.end(); it++ )
    {
        (*it).ptr->stop();
        delete (*it).ptr;
    }
    tclTimers.clear();
}

void TclSession::setID(int id)
{
    kdebugf();
    sessID = id;
}

int TclSession::getID()
{
    kdebugf();
    return sessID;
}

Tcl_Interp* TclSession::tclInt()
{
    // kdebugf() was generating too much craps here
    return tcl_int;
}

TclEventHandler* TclSession::tclEvtHnd()
{
    kdebugf();
    return tclEventHandler;
}

TclConsole* TclSession::console()
{
    kdebugf();
    return Tcl_cons;
}

QPopupMenu* TclSession::menu()
{
    kdebugf();
    return tclMenu;
}

void TclSession::setGui()
{
    kdebugf();
    guiEnabled = TRUE;
}

bool TclSession::isGui()
{
    kdebugf();
    return guiEnabled;
}

void TclSession::hideMainWindow()
{
    kdebugf();
    hiddingLoop = new QTimer();
    connect(hiddingLoop, SIGNAL(timeout()), this, SLOT(hideMainWindowLoop()));
    hiddingLoop->start(5);
}

void TclSession::hideMainWindowLoop()
{
    if (Tk_IsMapped(Tk_MainWindow(tcl_int)))
    {
        Tk_UnmapWindow(Tk_MainWindow(tcl_int));
        hiddingLoop->stop();
        delete hiddingLoop;
    }
}

void TclSession::processOneEvent()
{
#ifdef USE_TK
    for ( uint i = 0; i < 5; i++ )
        Tk_DoOneEvent(TK_ALL_EVENTS|TK_DONT_WAIT);
#endif
#ifndef USE_TK
    for ( uint i = 0; i < 5; i++ )
        Tcl_DoOneEvent(TCL_ALL_EVENTS|TCL_DONT_WAIT);
#endif
}

// Implementation of TclTimer class

TclTimer::TclTimer(TclSession* s)
{
    kdebugf();
    connect(this, SIGNAL(timeout()), this, SLOT(exec()));
    tsess = s;
}

void TclTimer::setCode(QString str)
{
    kdebugf();
    code = str;
}

void TclTimer::remove()
{
    kdebugf();
    QValueList<timerID>::Iterator it;
    for ( it = tsess->tclTimers.begin(); it != tsess->tclTimers.end(); it++ )
    {
        if ((*it).id == cid)
        {
            tsess->tclTimers.remove(it);
            delete (*it).ptr;
            break;
        }
    }
}

void TclTimer::exec()
{
    /*
        kdebugf() is ommited, becouse of craps
        it was generating in loop case.
    */
    QCString utf = code.utf8();
    Tcl_Obj* result = Tcl_NewStringObj((const char* const)utf, utf.length());
    int ret = Tcl_Eval(tsess->tclInt(), Tcl_GetString(result));
    if (ss)
        remove();

    if (ret == TCL_ERROR)
    {
        tsess->error(Tcl_GetStringResult(tsess->tclInt()));
        delete tsess;
    }
}

void TclTimer::kill()
{
    kdebugf();
    stop();
    remove();
}

void TclTimer::currIdent(int id)
{
    kdebugf();
    cid = id;
}

void TclTimer::setSingleShot(bool shot)
{
    kdebugf();
    ss = shot;
}

bool TclTimer::isSingleShot()
{
    kdebugf();
    return ss;
}

// Implementation of TclConsole (Konsola) class

TclConsole::TclConsole()
{
    kdebugf();
    setCaption(tr("TCL Console"));
    setMinimumSize(400, 200);
    resize(600, 400);

    // Results
    QFont *font = new QFont("misc-fixed");
    font->setPointSize(12);
    font->setBold(TRUE);

    txt = new TclTextEdit(this);
    txt->setReadOnly(TRUE);
    txt->setCurrentFont(*font);
    txt->setColor(QColor("gray"));
    txt->setPaper(QColor("black"));
    txt->setMinimumSize(300, 130);
    
    // Command line
    cmd = new TclLineEdit(this);
    cmd->setFont(*font);
    cmd->setPaletteForegroundColor(QColor("white"));
    cmd->setPaletteBackgroundColor(QColor("black"));
    cmd->setMinimumWidth(240);

    // Executing button
    QPushButton *btn = new QPushButton(this);
    btn->setText("Execute");
    
    // TCL session choosing box
    cbox = new QComboBox(FALSE, this);

    // Command line and button
    QGridLayout *g = new QGridLayout(this, 1, 1);
    g->addMultiCellWidget(txt, 0, 12, 0, 20);
    g->addMultiCellWidget(cmd, 13, 13, 0, 14);
    g->addMultiCellWidget(btn, 13, 13, 15, 16);
    g->addMultiCellWidget(cbox, 13, 13, 17, 20);

    // Command line and results
    connect(btn, SIGNAL(clicked()), this, SLOT(exec()));
    connect(cmd, SIGNAL(enter()), this, SLOT(exec()));
}

void TclConsole::mousePressEvent(QMouseEvent* e)
{
    kdebugf();
    getCmdLine()->setFocus();
}

TclLineEdit* TclConsole::getCmdLine()
{
    kdebugf();
    return cmd;
}

void TclConsole::exec()
{
    kdebugf();
    QString code = cmd->text(), file;
    cmd->clear();
    uint i = 0;
    while ( i < tclSess.count() && QStringList::split("/", tclSess[i]->scriptName()).last() != cbox->currentText() )
        i++;

    if (i == tclSess.count())
        i = 0;

    Tcl_Obj* result;
    QCString utf = code.utf8();
    result = Tcl_NewStringObj((const char* const)utf, utf.length());
    Tcl_Eval(tclSess[i]->tclInt(), Tcl_GetString(result));
    puts(QString::fromUtf8(Tcl_GetStringResult(tclSess[i]->tclInt())));
    cmd->histCnt = 0;
    cmd->history[0] = code;
    cmd->history.prepend("");
    uint max = cmd->consoleHistory + 1;
    while (cmd->history.count() > max)
    {
        QStringList::iterator it = cmd->history.end();
        it--;
        cmd->history.remove(it);
    }
}

void TclConsole::puts(QString str)
{
    kdebugf();
    txt->setReadOnly(FALSE);
    txt->append(str);
    txt->setReadOnly(TRUE);
}

QComboBox* TclConsole::context()
{
    kdebugf();
    return cbox;
}

// Implementation of TclListBox class

TclListBox::TclListBox(bool mesg)
{
    kdebugf();
    setMinimumWidth(400);
    setMinimumHeight(300);

    QPushButton *btn_ok = new QPushButton(this);
    btn_ok->setText(tr("OK"));
    btn_ok->setDefault(FALSE);

    QPushButton *btn_ccl = new QPushButton(this);
    btn_ccl->setText(tr("Cancel"));

    QPushButton *btn_add = new QPushButton(this);
    btn_add->setText(tr("Add"));
    btn_add->setDefault(TRUE);

    QPushButton *btn_del = new QPushButton(this);
    btn_del->setText(tr("Delete"));

    toAdd = new QLineEdit(this);
    toAdd->setMinimumWidth(100);
    toAdd->setFocus();

    _list = new QListBox(this);
    _list->setMinimumWidth(400);
    _list->setMinimumHeight(200);

    QGridLayout *g = new QGridLayout(this, 1, 1);
    if (mesg)
    {
        QVBoxLayout *l = new QVBoxLayout(this);
        msg = new QLabel(this);
        l->addWidget(msg);
        l->setMargin(10);
        g->addMultiCellLayout(l, 0, 0, 0, 20);
        g->addMultiCellWidget(toAdd, 1, 1, 0, 10);
        g->addMultiCellWidget(btn_add, 1, 1, 11, 14);
        g->addMultiCellWidget(btn_del, 1, 1, 16, 19);
        g->addMultiCellWidget(_list, 2, 22, 0, 20);
        g->addMultiCellWidget(btn_ok, 24, 24, 1, 5);
        g->addMultiCellWidget(btn_ccl, 24, 24, 14, 19);
    }
    else
    {
        g->addMultiCellWidget(toAdd, 0, 0, 0, 10);
        g->addMultiCellWidget(btn_add, 0, 0, 11, 14);
        g->addMultiCellWidget(btn_del, 0, 0, 16, 19);
        g->addMultiCellWidget(_list, 1, 21, 0, 20);
        g->addMultiCellWidget(btn_ok, 23, 23, 1, 5);
        g->addMultiCellWidget(btn_ccl, 23, 23, 14, 19);
    }

    connect(btn_ok, SIGNAL(clicked()), this, SLOT(ok()));
    connect(btn_ccl, SIGNAL(clicked()), this, SLOT(cancel()));
    connect(btn_add, SIGNAL(clicked()), this, SLOT(add()));
    connect(btn_del, SIGNAL(clicked()), this, SLOT(del()));
}

void TclListBox::add()
{
    kdebugf();
    _list->insertItem(toAdd->text());
    toAdd->clear();
}

void TclListBox::del()
{
    kdebugf();
    _list->removeItem(_list->currentItem());
}

void TclListBox::ok()
{
    kdebugf();
    accept();
}

void TclListBox::cancel()
{
    kdebugf();
    reject();
}

void TclListBox::setMsg(QString mesg)
{
    kdebugf();
    msg->setText(mesg);
}

QListBox* TclListBox::list()
{
    kdebugf();
    return _list;
}

void TclListBox::keyPressEvent(QKeyEvent* e)
{
    kdebugf();
    switch (e->key())
    {
        case Qt::Key_Delete:
            del();
            break;
           case Qt::Key_Return:
         case Qt::Key_Enter:
            add();
            break;
    }
}

// Implementation of TclLoadDialog class
TclLoadDialog::TclLoadDialog()
{
    kdebugf();
    setCaption(tr("TCL Scripts"));
    setMinimumWidth(250);
    setMinimumHeight(300);
    QFont labFont("helvetica", 12, QFont::Bold);

    // Close
    QPushButton* btn_close = new QPushButton(this);
    btn_close->setText(tr("Close"));
    btn_close->setDefault(FALSE);

    // Load
    QPushButton* btn_load = new QPushButton(this);
    btn_load->setText(tr("Load"));
    btn_load->setDefault(TRUE);

    // Unload
    QPushButton *btn_unload = new QPushButton(this);
    btn_unload->setText(tr("Unload"));

    // Loaded
    QLabel* loaded_label = new QLabel(this);
    loaded_label->setText(tr("Loaded scripts"));
    loaded_label->setFont(labFont);
    loaded_list = new QListBox(this);
    loaded_list->setMinimumWidth(150);
    loaded_list->setMinimumHeight(200);
    loaded_list->setAutoScroll(TRUE);

    // Local
    QLabel* local_label = new QLabel(this);
    local_label->setText(tr("Local scripts"));
    local_label->setFont(labFont);
    local_list = new QListBox(this);
    local_list->setMinimumWidth(150);
    local_list->setMinimumHeight(90);
    local_list->setAutoScroll(TRUE);

    // Global
    QLabel* global_label = new QLabel(this);
    global_label->setText(tr("Global scripts"));
    global_label->setFont(labFont);
    global_list = new QListBox(this);
    global_list->setMinimumWidth(150);
    global_list->setMinimumHeight(90);
    global_list->setAutoScroll(TRUE);

    // Others
    QPushButton* btn_others = new QPushButton(this);
    btn_others->setText(tr("Other scripts"));

    // Others
    QPushButton* btn_reload = new QPushButton(this);
    btn_reload->setText(tr("Reload"));

    // Layout
    QGridLayout *g = new QGridLayout(this, 1, 1);

    g->addMultiCellWidget(btn_load, 0, 0, 0, 9);
    g->addMultiCellWidget(btn_unload, 0, 0, 11, 20);
 
    g->addMultiCellWidget(local_label, 1, 1, 0, 9);
    g->addMultiCellWidget(local_list, 2, 10, 0, 9);
    g->addMultiCellWidget(global_label, 11, 11, 0, 9);
    g->addMultiCellWidget(global_list, 12, 19, 0, 9);
    g->addMultiCellWidget(loaded_label, 1, 1, 11, 20);
    g->addMultiCellWidget(loaded_list, 2, 19, 11, 20);
 
    g->addMultiCellWidget(btn_others, 20, 20, 0, 9);
    g->addMultiCellWidget(btn_reload, 20, 20, 11, 20);
    g->addMultiCellWidget(btn_close, 21, 21, 0, 20);

    resize(250, 350);
    
    // Slots    
    connect(btn_close, SIGNAL(clicked()), this, SLOT(done()));
    connect(btn_others, SIGNAL(clicked()), this, SLOT(others()));
    connect(btn_load, SIGNAL(clicked()), this, SLOT(load()));
    connect(btn_unload, SIGNAL(clicked()), this, SLOT(unload()));
    connect(btn_reload, SIGNAL(clicked()), this, SLOT(reload()));

    connect(local_list, SIGNAL(doubleClicked(QListBoxItem*)), this, SLOT(loadLocal(QListBoxItem*)));
    connect(global_list, SIGNAL(doubleClicked(QListBoxItem*)), this, SLOT(loadGlobal(QListBoxItem*)));
    connect(loaded_list, SIGNAL(doubleClicked(QListBoxItem*)), this, SLOT(unload(QListBoxItem*)));

    // Paths
    localPath = ggPath("");
    globalPath = dataPath("kadu/modules/data/tcl_scripting/scripts/");

    // Making local and global scripts list
    QStringList locals = QDir(localPath).entryList(),
                globals = QDir(globalPath).entryList();
    for ( uint i = 0; i < locals.count(); i++ )
    {
        if (locals[i].find(".tcl", FALSE) > -1 && locals[i].compare("default.tcl"))
        {
            local_list->insertItem(locals[i]);
        }
    }
    for ( uint i = 0; i < globals.count(); i++ )
    {
        if (globals[i].find(".tcl", FALSE) > -1)
        {
            global_list->insertItem(globals[i]);
        }
    }
}

void TclLoadDialog::load(int list_type)
{
    kdebugf();
    QString item, file;
    if (list_type == 0)
    {
        if (local_list->currentItem() > -1)
        {
            item = local_list->currentText();
            file = localPath + item;
            list_type = 1;
        }
        else if (global_list->currentItem() > -1)
        {
            item = global_list->currentText();
            file = globalPath + item;
            list_type = 2;
        }
        else
        {
            return;
        }
    }
    else
    {
        switch (list_type)
        {
            case 1:
                item = local_list->currentText();
                file = localPath + item;
                break;
            case 2:
                item = global_list->currentText();
                file = globalPath + item;
                break;
        }
    }

    if (loaded(item))
    {
        QMessageBox::critical( 0, tr("Error"), tr("Script %1 is already loaded.").arg(item),
            QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);

        return;
    }

    TclSession* tsess = new TclSession(FALSE);

    bool ret = tsess->load(file);

    if (ret == TCL_ERROR)
    {
        tsess->error(QString::fromUtf8(Tcl_GetStringResult(tsess->tclInt())), TRUE);
        delete tsess;
    }
    else
    {
        loaded_list->insertItem(item);
        switch (list_type)
        {
            case 1:
            {
                local_list->removeItem(local_list->currentItem());
                break;
            }
            case 2:
            {
                global_list->removeItem(global_list->currentItem());
                break;
            }
        }
    }
}

void TclLoadDialog::loadLocal(QListBoxItem*)
{
    kdebugf();
    load(1);
}

void TclLoadDialog::loadGlobal(QListBoxItem*)
{
    kdebugf();
    load(2);
}

void TclLoadDialog::unload(QListBoxItem* it)
{
    kdebugf();
    if (loaded_list->currentItem() == -1)
        return;

    QString item = loaded_list->currentText(), script;
    for ( TclSessions::Iterator it = tclSess.begin(); it != tclSess.end(); it++ )
    {
        script = QStringList::split("/", it.data()->scriptName() ).last();
        if (script == item)
        {
            loaded_list->removeItem(loaded_list->currentItem());
            delete it.data();
            break;
        }
    }
    QFile localScripts(QString("%1%2").arg(localPath).arg(item));
    QFile globalScripts(QString("%1%2").arg(globalPath).arg(item));
    if (localScripts.exists())
    {
        bool sw = FALSE;
        for ( uint i = 0; i < local_list->count(); i++ )
        {
            if (!local_list->text(i).compare(item))
            {
                sw = TRUE;
                break;
            }
        }
        if (sw)
        {
            if (globalScripts.exists())
            {
                global_list->insertItem(item);
                return;
            }
        }
        local_list->insertItem(item);
    }
    else if (globalScripts.exists())
    {
        global_list->insertItem(item);
    }
}

void TclLoadDialog::reload()
{
    kdebugf();
    if (!loaded_list->selectedItem())
        return;

    QString item = loaded_list->currentText(), script, file;
    for ( TclSessions::Iterator it = tclSess.begin(); it != tclSess.end(); it++ )
    {
        file = it.data()->scriptName();
        script = QStringList::split("/", file).last();
        if (script == item)
        {
            delete it.data();

            TclSession* tsess = new TclSession(FALSE);
            bool ret = tsess->load(file);
            if (ret == TCL_ERROR)
            {
                loaded_list->removeItem(loaded_list->currentItem());
                QFile localScripts(QString("%1%2").arg(localPath).arg(item));
                QFile globalScripts(QString("%1%2").arg(globalPath).arg(item));
                if (localScripts.exists())
                {
                    bool sw = FALSE;
                    for ( uint i = 0; i < local_list->count(); i++ )
                    {
                        if (!local_list->text(i).compare(item))
                        {
                            sw = TRUE;
                            break;
                        }
                    }
                    if (sw)
                    {
                        if (globalScripts.exists())
                        {
                            global_list->insertItem(item);
                            return;
                        }
                    }
                    local_list->insertItem(item);
                }
                else if (globalScripts.exists())
                {
                    global_list->insertItem(item);
                }
            }
            break;
        }
    }
}

void TclLoadDialog::done()
{
    kdebugf();
    close();
}

void TclLoadDialog::others()
{
    kdebugf();
    QFileDialog *fd = new QFileDialog();
    fd->setCaption(tr("Choose file to load"));
    fd->setMode(QFileDialog::ExistingFile);
    fd->setFilter(tr("TCL scripts (%1)").arg("*.tcl *.TCL"));
    fd->show();
    if ( fd->exec() == QDialog::Accepted )
    {
        QString file = fd->selectedFile(), item = QStringList::split("/", file).last();
        if (loaded(item))
        {
            QMessageBox::critical( 0, tr("Error"), tr("Script %1 is already loaded.").arg(item),
                QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);

            return;
        }

        TclSession* tsess = new TclSession(FALSE);

        bool ret = tsess->load(file.latin1());

        if (ret == TCL_ERROR)
        {
            tsess->error(QString::fromUtf8(Tcl_GetStringResult(tsess->tclInt())), TRUE);
            delete tsess;
        }
        else
        {
            loaded_list->insertItem(QStringList::split("/", file ).last());
        }
    }
}

QListBox* TclLoadDialog::list()
{
    kdebugf();
    return loaded_list;
}

QListBox* TclLoadDialog::lList()
{
    kdebugf();
    return local_list;
}

QListBox* TclLoadDialog::gList()
{
    kdebugf();
    return global_list;
}

bool TclLoadDialog::loaded(QString script)
{
    kdebugf();
    bool l = FALSE;
    for ( uint i = 0; i < loaded_list->count(); i++ )
    {
        if (loaded_list->text(i) == script)
        {
            l = TRUE;
            break;
        }
    }
    return l;
}

void TclLoadDialog::keyPressEvent(QKeyEvent* e)
{
    kdebugf();
    switch (e->key())
    {
        case Qt::Key_Delete:
            unload();
            break;
        case Qt::Key_Insert:
            load();
            break;
    }
}

// Implementation of TclLineEdit class

TclLineEdit::TclLineEdit(QWidget *parent, const char *name) : QLineEdit(parent, name)
{
    kdebugf();
    consoleHistory = 10;
}

void TclLineEdit::keyPressEvent(QKeyEvent *e)
{
    /*
        kdebugf() is ommited, becouse of craps it was generating.
    */
    QString str;

    if (e->state() == Qt::ControlButton)
    {
        switch (e->key())
        {
            case Qt::Key_C:
                copy();
                break;
            case Qt::Key_X:
                cut();
                break;
            case Qt::Key_V:
                paste();
                break;
            default:
                e->ignore();
        }
    }
    else if (e->state() == Qt::ShiftButton)
    {
        switch (e->key())
        {
            case Qt::Key_Left:
                cursorBackward(TRUE);
                break;
            case Qt::Key_Right:
                cursorForward(TRUE);
                break;
            default:
                insert(e->text());
        }
    }
    else if (e->state() == Qt::KeyButtonMask)
    {
        e->ignore();
    }
    else
    {
        switch (e->key())
        {
            case Qt::Key_Up:
                historyUp();
                break;
            case Qt::Key_Down:
                historyDown();
                break;
            case Qt::Key_PageUp:
                e->ignore();
                break;
            case Qt::Key_PageDown:
                e->ignore();
                break;
            case Qt::Key_Left:
                cursorBackward(FALSE);
                break;
            case Qt::Key_Right:
                cursorForward(FALSE);
                break;
            case Qt::Key_Backspace:
                if (e->state() == Qt::ShiftButton)
                {
                    cursorWordBackward(TRUE);
                }
                backspace();
                break;
            case Qt::Key_Delete:
                if (e->state() == Qt::ShiftButton)
                {
                    cursorWordForward(TRUE);
                }
                del();
                break;
            case Qt::Key_Home:
                home(FALSE);
                break;
            case Qt::Key_End:
                end(FALSE);
                break;
            case Qt::Key_Return:
            case Qt::Key_Enter:
                emit enter();
                break;
            default:
                insert(e->text());
        }
    }
}

void TclLineEdit::historyUp()
{
    kdebugf();
    if (histCnt < consoleHistory && histCnt < history.count())
    {
        histCnt++;
        setText(history[histCnt]);
    }
}

void TclLineEdit::historyDown()
{
    kdebugf();
    if (histCnt > 0)
    {
        histCnt--;
        setText(history[histCnt]);
    }
}

// Implemention of TclTextEdit class
TclTextEdit::TclTextEdit(QWidget *parent, const char *name) : QTextEdit(parent, name)
{
    kdebugf();
}

void TclTextEdit::focusInEvent(QFocusEvent* e)
{
    kdebugf();
    tclSess[0]->console()->getCmdLine()->setFocus();
}

// Implementation of TclMisc class
QString TclMisc::GGStatusToTclStatus(int status)
{
    switch (GG_S(status))
    {
        case GG_STATUS_NOT_AVAIL:
            return "notavailable";
            break;
        case GG_STATUS_AVAIL:
            return "available";
            break;
        case GG_STATUS_BUSY:
            return "busy";
            break;
        case GG_STATUS_AVAIL_DESCR:
            return "available+description";
            break;
        case GG_STATUS_BUSY_DESCR:
            return "busy+description";
            break;
        case GG_STATUS_NOT_AVAIL_DESCR:
            return "notavailable+description";
            break;
        case GG_STATUS_INVISIBLE:
        case GG_STATUS_INVISIBLE2:
            return "invisible";
            break;
        case GG_STATUS_INVISIBLE_DESCR:
            return "invisible+description";
            break;
        case GG_STATUS_BLOCKED:
            return "blocked";
            break;
        default:
            return "";
    }
}

int TclMisc::TclStatusToGGStatus(QString status)
{
    QStringList stats = QStringList::split(" ", "notavailable available busy available+description busy+description notavailable+description invisible invisible+description");
    int id = stats.findIndex(status);
    switch (id)
    {
        case 0:
            return GG_STATUS_NOT_AVAIL;
            break;
        case 1:
            return GG_STATUS_AVAIL;
            break;
        case 2:
            return GG_STATUS_BUSY;
            break;
        case 3:
            return GG_STATUS_AVAIL_DESCR;
            break;
        case 4:
            return GG_STATUS_BUSY_DESCR;
            break;
        case 5:
            return GG_STATUS_NOT_AVAIL_DESCR;
            break;
        case 6:
            return GG_STATUS_INVISIBLE;
            break;
        case 7:
            return GG_STATUS_INVISIBLE_DESCR;
            break;
        default:
            return -1;
    }
}

void TclMisc::getKey(QKeyEvent* e, QString& keystr, QString& modstr)
{
    switch (e->state())
    {
        case Qt::ShiftButton:
            modstr = "shift";
            break;
        case Qt::ControlButton:
            modstr = "control";
            break;
        case Qt::AltButton:
            modstr = "alt";
            break;
        case Qt::MetaButton:
            modstr = "meta";
            break;
        case Qt::Keypad:
            modstr = "keypad";
            break;
        default:
            break;
    }

    int key = e->key();
    if (keystr[0].isPrint())
        return;

    switch (key)
    {
        case Qt::Key_Super_L:
            keystr = "win_l";
            break;
        case Qt::Key_Super_R:
            keystr = "win_r";
            break;
        case Qt::Key_Shift:
            keystr = "shift";
            break;
        case Qt::Key_Control:
            keystr = "control";
            break;
        case Qt::Key_Meta:
            keystr = "meta";
            break;
        case Qt::Key_Alt:
            keystr = "alt";
            break;
        case Qt::Key_Escape:
            keystr = "escape";
            break;
        case Qt::Key_Tab:
            keystr = "tab";
            break;
        case Qt::Key_Backspace:
            keystr = "backspace";
            break;
        case Qt::Key_Return:
        case Qt::Key_Enter:
            keystr = "enter";
            break;
        case Qt::Key_Insert:
            keystr = "insert";
            break;
        case Qt::Key_Delete:
            keystr = "delete";
            break;
        case Qt::Key_Pause:
            keystr = "pause";
            break;
        case Qt::Key_Print:
            keystr = "print";
            break;
        case Qt::Key_SysReq:
            keystr = "sysreq";
            break;
        case Qt::Key_Home:
            keystr = "home";
            break;
        case Qt::Key_End:
            keystr = "end";
            break;
        case Qt::Key_Left:
            keystr = "left";
            break;
        case Qt::Key_Up:
            keystr = "up";
            break;
        case Qt::Key_Right:
            keystr = "right";
            break;
        case Qt::Key_Down:
            keystr = "down";
            break;
        case Qt::Key_Prior:
            keystr = "pgup";
            break;
        case Qt::Key_Next:
            keystr = "pgdown";
            break;
        case Qt::Key_CapsLock:
            keystr = "capslock";
            break;
        case Qt::Key_NumLock:
            keystr = "numlock";
            break;
        case Qt::Key_ScrollLock:
            keystr = "scrolllock";
            break;
        case Qt::Key_Clear:
            keystr = "clear";
            break;
        case Qt::Key_F1:
            keystr = "f1";
            break;
        case Qt::Key_F2:
            keystr = "f2";
            break;
        case Qt::Key_F3:
            keystr = "f3";
            break;
        case Qt::Key_F4:
            keystr = "f4";
            break;
        case Qt::Key_F5:
            keystr = "f5";
            break;
        case Qt::Key_F6:
            keystr = "f6";
            break;
        case Qt::Key_F7:
            keystr = "f7";
            break;
        case Qt::Key_F8:
            keystr = "f8";
            break;
        case Qt::Key_F9:
            keystr = "f9";
            break;
        case Qt::Key_F10:
            keystr = "f10";
            break;
        case Qt::Key_F11:
            keystr = "f11";
            break;
        case Qt::Key_F12:
            keystr = "f12";
            break;
        case Qt::Key_Menu:
            keystr = "menu";
            break;
        default:
            keystr = "unknown";
    }
}
