#ifndef TCL_H
#define TCL_H

//#include "config_tcl_scripting.h"
#define USE_TK

#include <tcl.h>

#include <qobject.h>
#include <qtimer.h>
#include <qlineedit.h>
#include <qdialog.h>
#include <qlistbox.h>
#include <qtextedit.h>
#include <qstring.h>
#include <qstringlist.h>
#include <qvaluelist.h>
#include <qmap.h>
#include <qfiledialog.h>
#include <qpopupmenu.h>
#include <qmessagebox.h>

#include "kadu.h"
#include "userinfo.h"
#include "gadu.h"
#include "debug.h"
#include "chat.h"

class TclSession;

class TclEventHandler : public QObject
{
    Q_OBJECT

    public:
        TclEventHandler(TclSession*);
        ~TclEventHandler();
        void unloadModule();
        void checkForNext(gg_pubdir50_t, int);

    private slots:
        void connected();
        void connectionFailed(int);
        void connectionBroken();
        void disconnected();
        void userStatusChanging(const UinType, const unsigned int, const unsigned int);
        void userStatusChanged(struct gg_event*);
        void userlistReceived(struct gg_event *);
        void messageReceived(int, UinsList, QCString&, time_t, QByteArray&);
        void chatReceived0(UinsList senders, const QString& msg, time_t time, bool& grab);
        void chatReceived1(UinsList senders, const QString& msg, time_t time, bool& grab);
        void chatReceived2(UinsList senders, const QString& msg, time_t time);
        void ackReceived(int seq);
        void dccConnectionReceived(const UserListElement& sender);
        void pubdirReplyReceived(gg_pubdir50_t res);
        void userlistReplyReceived(char type, char *reply);
        void chatCreated(const UinsList& senders);
        void chatDestroyed(const UinsList&);
        void userlist_modified();
        void userlist_userAdded(const UserListElement&);
        void userlist_dnsNameReady(UinType);
        void chatMsgSendReq(Chat*);
        void chatMsgSent(Chat*);
        void chatCreated_Watch(const UinsList&);
        void chatKeyPressed(QKeyEvent*, CustomInput*);
        void searchInPubDirFinished(int);
        void changingStatus(int, QString, bool&);
        void outputMessageFiltering(const UinsList&, QCString&, bool&);

    public slots:
        void msgBoxCancel();

    signals:
        void searchingFinished(int);

    private:
        TclSession* tsess;
        void manageReadyChats(Chat*, bool);
};

class TclTimer : public QTimer
{
    Q_OBJECT

    public:
        TclTimer(TclSession*);
        void setCode(QString);
        void currIdent(int);
        void remove();
        void kill();
        void setSingleShot(bool);
        bool isSingleShot();

    private slots:
        void exec();

    private:
        int cid;
        QString code;
        bool ss;
        TclSession* tsess;
};

class TclLineEdit : public QLineEdit
{
    Q_OBJECT

    public:
        TclLineEdit(QWidget *parent = 0, const char *name = 0);
        void keyPressEvent(QKeyEvent*);
        QStringList history;
        uint consoleHistory;
        uint histCnt;

    private slots:
        void historyUp();
        void historyDown();

    signals:
        void enter();
};

class TclTextEdit : public QTextEdit
{
    public:
        TclTextEdit(QWidget *parent = 0, const char *name = 0);
        void focusInEvent(QFocusEvent*);
};

class TclConsole : public QWidget
{
    Q_OBJECT

    public:
        TclConsole();
        TclLineEdit* getCmdLine();
        QComboBox* context();
        void mousePressEvent(QMouseEvent*);

    private:
        TclLineEdit* cmd;
        void puts(QString);
        QTextEdit* txt;
        QComboBox* cbox;
    
    private slots:
        void exec();
};

class TclListBox : public QDialog
{
    Q_OBJECT

    public:
        TclListBox(bool mesg = FALSE);
        void keyPressEvent(QKeyEvent*);
        void setMsg(QString);
        QListBox* list();

    private:
        QLineEdit *toAdd;
        QListBox *_list;
        QLabel *msg;

    private slots:
        void add();
        void del();
        void ok();
        void cancel();
};

class TclLoadDialog : public QDialog
{
    Q_OBJECT

    public:
        TclLoadDialog();
        void keyPressEvent(QKeyEvent*);
        void setMsg(QString);
        QListBox* list();
        QListBox* lList();
        QListBox* gList();
        bool loaded(QString);

    private:
        QListBox *loaded_list;
        QListBox *global_list;
        QListBox *local_list;
        QString localPath;
        QString globalPath;

    private slots:
        void done();
        void load(int list_type = 0);
        void loadLocal(QListBoxItem*);
        void loadGlobal(QListBoxItem*);
        void unload(QListBoxItem* it = NULL);
        void others();
        void reload();
};

struct timerID
{
    int id;
    TclTimer *ptr;
};

struct TclPopup
{
    QPopupMenu *menu;
    int id;
    QMap<int,QString> cmd;
};

class TclSession : public QObject
{
    Q_OBJECT

    private:
        typedef QMap<QString,TclPopup> TclPopups;
        typedef QMap<QString,QString> TclEvents;
        typedef QValueList<timerID> TclTimers;
        typedef QMap<int,QString> TclPopupCmds;
        typedef QMap<int,QString> TclMenuCmds;
        typedef QMap<QString,int> TclMenuBar;

        bool start();
        void stop();
        void addTclCommands();
        int load();

        TclConsole* Tcl_cons;
        Tcl_Interp* tcl_int;
        TclEventHandler* tclEventHandler;
        QPopupMenu* tclMenu;
        int sessID;
        int* idPtr;
        int menu_id;
        bool isMain;
        bool tclrun;
        QString scriptFile;
        bool guiEnabled;
        QTimer *hiddingLoop;

    public:
        TclSession::TclSession(bool main = FALSE);
        TclSession::~TclSession();
        void addMenu();
        void delMenu();
        void menuItemSetEnabled(int idx, bool enabled);
        int load(QString file);
        bool isrun();
        void error(QString cmd, bool line = FALSE);
        QString scriptName();
        void killAllTimers();
        void setID(int);
        int getID();
        Tcl_Interp* tclInt();
        TclEventHandler* tclEvtHnd();
        TclConsole* console();
        QPopupMenu* menu();
        void setGui();
        bool isGui();
        void hideMainWindow();
        
        TclPopups tclPopups;
        TclEvents tclEventList;
        TclTimers tclTimers;
        TclPopupCmds tclPopupCmds;
        TclMenuCmds tclMenuCmds;
        TclMenuBar tclMenuBar;

    public slots:
        void openLoadDialog();
        void openConsole();
        void handlePopup(int id);
        void handleUserMenu(int id);
        void hideTclMenu();
        void showTclMenu();
        void aboutModule();
        void processOneEvent();
        void hideMainWindowLoop();
};

class TclMisc
{
    public:
        static QString GGStatusToTclStatus(int);
        static int TclStatusToGGStatus(QString);
        static void getKey(QKeyEvent*, QString&, QString&);
};

typedef QMap<int, TclSession*> TclSessions;
typedef QMap<int, int> PubDirID;

extern TclSessions tclSess;
extern TclLoadDialog *loadDialog;
extern QTimer* events_loop;
extern PubDirID tclPubDirList;

#endif
